@@ -128,3 +128,34 @@ class BaseCtx(
         if not options:
             options = cls.options_cls
         return getattr(options, option_name)
+
+
+class Code:
+    """A building block for generating code, which is composed of multiple
+    statements."""
+
+    def __init__(self):
+        self.lines_info = []
+        self.indent_level = 0
+
+    def add_line(self, line: str, next_line_indent_incr: int):
+        self.lines_info.append((self.indent_level, line))
+        self.indent_level += next_line_indent_incr
+        if self.indent_level < 0:
+            raise AssertionError("negative indentation level")
+
+    def add_code(self, code: "Code"):
+        for indent_level, line in code.lines_info:
+            self.lines_info.append((indent_level + self.indent_level, line))
+
+    def incr_indent_level(self, incr: int):
+        self.indent_level += incr
+
+    def has_lines(self):
+        return bool(self.lines_info)
+
+    def to_string(self, base_indent_level: int, single_indent: str = "    "):
+        return "\n".join(
+            f"{single_indent * (base_indent_level + indent_level)}{line}"
+            for indent_level, line in self.lines_info
+        )
